"""
A LC-3 simulator and unit tester, backed by lc3-ensemble.

All of the contents of this package can be found 
in the subpackages `ensemble_test.core` and `ensemble_test.autograder`.

See `ensemble_test.core` for details about simulation,
and `ensemble_test.autograder` for details about unit testing.
"""
